@@ -1,5 +1,5 @@
 /*
-Copyright 2018 the Velero contributors.
+Copyright 2020 the Velero contributors.
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -23,6 +23,9 @@ import (
 	"time"
 
 	jsonpatch "github.com/evanphx/json-patch"
+	snapshotterClientSet "github.com/kubernetes-csi/external-snapshotter/v2/pkg/client/clientset/versioned"
+	snapshotter "github.com/kubernetes-csi/external-snapshotter/v2/pkg/client/clientset/versioned/typed/volumesnapshot/v1beta1"
+	snapshotv1beta1listers "github.com/kubernetes-csi/external-snapshotter/v2/pkg/client/listers/volumesnapshot/v1beta1"
 	"github.com/pkg/errors"
 	"github.com/sirupsen/logrus"
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
@@ -33,18 +36,24 @@ import (
 	kubeerrs "k8s.io/apimachinery/pkg/util/errors"
 	"k8s.io/client-go/tools/cache"
 
-	v1 "github.com/vmware-tanzu/velero/pkg/apis/velero/v1"
+	"github.com/vmware-tanzu/velero/internal/delete"
+	velerov1api "github.com/vmware-tanzu/velero/pkg/apis/velero/v1"
 	pkgbackup "github.com/vmware-tanzu/velero/pkg/backup"
+	"github.com/vmware-tanzu/velero/pkg/discovery"
+	"github.com/vmware-tanzu/velero/pkg/features"
 	velerov1client "github.com/vmware-tanzu/velero/pkg/generated/clientset/versioned/typed/velero/v1"
-	informers "github.com/vmware-tanzu/velero/pkg/generated/informers/externalversions/velero/v1"
-	listers "github.com/vmware-tanzu/velero/pkg/generated/listers/velero/v1"
+	velerov1informers "github.com/vmware-tanzu/velero/pkg/generated/informers/externalversions/velero/v1"
+	velerov1listers "github.com/vmware-tanzu/velero/pkg/generated/listers/velero/v1"
 	"github.com/vmware-tanzu/velero/pkg/label"
 	"github.com/vmware-tanzu/velero/pkg/metrics"
 	"github.com/vmware-tanzu/velero/pkg/persistence"
 	"github.com/vmware-tanzu/velero/pkg/plugin/clientmgmt"
 	"github.com/vmware-tanzu/velero/pkg/plugin/velero"
 	"github.com/vmware-tanzu/velero/pkg/restic"
+	"github.com/vmware-tanzu/velero/pkg/util/filesystem"
 	"github.com/vmware-tanzu/velero/pkg/util/kube"
+
+	"sigs.k8s.io/controller-runtime/pkg/client"
 )
 
 const resticTimeout = time.Minute
@@ -53,51 +62,63 @@ type backupDeletionController struct {
 	*genericController
 
 	deleteBackupRequestClient velerov1client.DeleteBackupRequestsGetter
-	deleteBackupRequestLister listers.DeleteBackupRequestLister
+	deleteBackupRequestLister velerov1listers.DeleteBackupRequestLister
 	backupClient              velerov1client.BackupsGetter
-	restoreLister             listers.RestoreLister
+	restoreLister             velerov1listers.RestoreLister
 	restoreClient             velerov1client.RestoresGetter
 	backupTracker             BackupTracker
 	resticMgr                 restic.RepositoryManager
-	podvolumeBackupLister     listers.PodVolumeBackupLister
-	backupLocationLister      listers.BackupStorageLocationLister
-	snapshotLocationLister    listers.VolumeSnapshotLocationLister
-	processRequestFunc        func(*v1.DeleteBackupRequest) error
+	podvolumeBackupLister     velerov1listers.PodVolumeBackupLister
+	kbClient                  client.Client
+	snapshotLocationLister    velerov1listers.VolumeSnapshotLocationLister
+	csiSnapshotLister         snapshotv1beta1listers.VolumeSnapshotLister
+	csiSnapshotContentLister  snapshotv1beta1listers.VolumeSnapshotContentLister
+	csiSnapshotClient         *snapshotterClientSet.Clientset
+	processRequestFunc        func(*velerov1api.DeleteBackupRequest) error
 	clock                     clock.Clock
 	newPluginManager          func(logrus.FieldLogger) clientmgmt.Manager
-	newBackupStore            func(*v1.BackupStorageLocation, persistence.ObjectStoreGetter, logrus.FieldLogger) (persistence.BackupStore, error)
+	newBackupStore            func(*velerov1api.BackupStorageLocation, persistence.ObjectStoreGetter, logrus.FieldLogger) (persistence.BackupStore, error)
 	metrics                   *metrics.ServerMetrics
+	helper                    discovery.Helper
 }
 
 // NewBackupDeletionController creates a new backup deletion controller.
 func NewBackupDeletionController(
 	logger logrus.FieldLogger,
-	deleteBackupRequestInformer informers.DeleteBackupRequestInformer,
+	deleteBackupRequestInformer velerov1informers.DeleteBackupRequestInformer,
 	deleteBackupRequestClient velerov1client.DeleteBackupRequestsGetter,
 	backupClient velerov1client.BackupsGetter,
-	restoreInformer informers.RestoreInformer,
+	restoreLister velerov1listers.RestoreLister,
 	restoreClient velerov1client.RestoresGetter,
 	backupTracker BackupTracker,
 	resticMgr restic.RepositoryManager,
-	podvolumeBackupInformer informers.PodVolumeBackupInformer,
-	backupLocationInformer informers.BackupStorageLocationInformer,
-	snapshotLocationInformer informers.VolumeSnapshotLocationInformer,
+	podvolumeBackupLister velerov1listers.PodVolumeBackupLister,
+	kbClient client.Client,
+	snapshotLocationLister velerov1listers.VolumeSnapshotLocationLister,
+	csiSnapshotLister snapshotv1beta1listers.VolumeSnapshotLister,
+	csiSnapshotContentLister snapshotv1beta1listers.VolumeSnapshotContentLister,
+	csiSnapshotClient *snapshotterClientSet.Clientset,
 	newPluginManager func(logrus.FieldLogger) clientmgmt.Manager,
 	metrics *metrics.ServerMetrics,
+	helper discovery.Helper,
 ) Interface {
 	c := &backupDeletionController{
 		genericController:         newGenericController("backup-deletion", logger),
 		deleteBackupRequestClient: deleteBackupRequestClient,
 		deleteBackupRequestLister: deleteBackupRequestInformer.Lister(),
 		backupClient:              backupClient,
-		restoreLister:             restoreInformer.Lister(),
+		restoreLister:             restoreLister,
 		restoreClient:             restoreClient,
 		backupTracker:             backupTracker,
 		resticMgr:                 resticMgr,
-		podvolumeBackupLister:     podvolumeBackupInformer.Lister(),
-		backupLocationLister:      backupLocationInformer.Lister(),
-		snapshotLocationLister:    snapshotLocationInformer.Lister(),
+		podvolumeBackupLister:     podvolumeBackupLister,
+		kbClient:                  kbClient,
+		snapshotLocationLister:    snapshotLocationLister,
+		csiSnapshotLister:         csiSnapshotLister,
+		csiSnapshotContentLister:  csiSnapshotContentLister,
+		csiSnapshotClient:         csiSnapshotClient,
 		metrics:                   metrics,
+		helper:                    helper,
 		// use variables to refer to these functions so they can be
 		// replaced with fakes for testing.
 		newPluginManager: newPluginManager,
@@ -107,14 +128,6 @@ func NewBackupDeletionController(
 	}
 
 	c.syncHandler = c.processQueueItem
-	c.cacheSyncWaiters = append(
-		c.cacheSyncWaiters,
-		deleteBackupRequestInformer.Informer().HasSynced,
-		restoreInformer.Informer().HasSynced,
-		podvolumeBackupInformer.Informer().HasSynced,
-		backupLocationInformer.Informer().HasSynced,
-		snapshotLocationInformer.Informer().HasSynced,
-	)
 	c.processRequestFunc = c.processRequest
 
 	deleteBackupRequestInformer.Informer().AddEventHandler(
@@ -148,7 +161,7 @@ func (c *backupDeletionController) processQueueItem(key string) error {
 	}
 
 	switch req.Status.Phase {
-	case v1.DeleteBackupRequestPhaseProcessed:
+	case velerov1api.DeleteBackupRequestPhaseProcessed:
 		// Don't do anything because it's already been processed
 	default:
 		// Don't mutate the shared cache
@@ -159,7 +172,7 @@ func (c *backupDeletionController) processQueueItem(key string) error {
 	return nil
 }
 
-func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) error {
+func (c *backupDeletionController) processRequest(req *velerov1api.DeleteBackupRequest) error {
 	log := c.logger.WithFields(logrus.Fields{
 		"namespace": req.Namespace,
 		"name":      req.Name,
@@ -170,8 +183,8 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 
 	// Make sure we have the backup name
 	if req.Spec.BackupName == "" {
-		_, err = c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-			r.Status.Phase = v1.DeleteBackupRequestPhaseProcessed
+		_, err = c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+			r.Status.Phase = velerov1api.DeleteBackupRequestPhaseProcessed
 			r.Status.Errors = []string{"spec.backupName is required"}
 		})
 		return err
@@ -185,8 +198,8 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 
 	// Don't allow deleting an in-progress backup
 	if c.backupTracker.Contains(req.Namespace, req.Spec.BackupName) {
-		_, err = c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-			r.Status.Phase = v1.DeleteBackupRequestPhaseProcessed
+		_, err = c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+			r.Status.Phase = velerov1api.DeleteBackupRequestPhaseProcessed
 			r.Status.Errors = []string{"backup is still in progress"}
 		})
 
@@ -194,11 +207,11 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 	}
 
 	// Get the backup we're trying to delete
-	backup, err := c.backupClient.Backups(req.Namespace).Get(req.Spec.BackupName, metav1.GetOptions{})
+	backup, err := c.backupClient.Backups(req.Namespace).Get(context.TODO(), req.Spec.BackupName, metav1.GetOptions{})
 	if apierrors.IsNotFound(err) {
 		// Couldn't find backup - update status to Processed and record the not-found error
-		req, err = c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-			r.Status.Phase = v1.DeleteBackupRequestPhaseProcessed
+		req, err = c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+			r.Status.Phase = velerov1api.DeleteBackupRequestPhaseProcessed
 			r.Status.Errors = []string{"backup not found"}
 		})
 
@@ -209,21 +222,24 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 	}
 
 	// Don't allow deleting backups in read-only storage locations
-	location, err := c.backupLocationLister.BackupStorageLocations(backup.Namespace).Get(backup.Spec.StorageLocation)
-	if apierrors.IsNotFound(err) {
-		_, err := c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-			r.Status.Phase = v1.DeleteBackupRequestPhaseProcessed
-			r.Status.Errors = append(r.Status.Errors, fmt.Sprintf("backup storage location %s not found", backup.Spec.StorageLocation))
-		})
-		return err
-	}
-	if err != nil {
+	location := &velerov1api.BackupStorageLocation{}
+	if err := c.kbClient.Get(context.Background(), client.ObjectKey{
+		Namespace: backup.Namespace,
+		Name:      backup.Spec.StorageLocation,
+	}, location); err != nil {
+		if apierrors.IsNotFound(err) {
+			_, err := c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+				r.Status.Phase = velerov1api.DeleteBackupRequestPhaseProcessed
+				r.Status.Errors = append(r.Status.Errors, fmt.Sprintf("backup storage location %s not found", backup.Spec.StorageLocation))
+			})
+			return err
+		}
 		return errors.Wrap(err, "error getting backup storage location")
 	}
 
-	if location.Spec.AccessMode == v1.BackupStorageLocationAccessModeReadOnly {
-		_, err := c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-			r.Status.Phase = v1.DeleteBackupRequestPhaseProcessed
+	if location.Spec.AccessMode == velerov1api.BackupStorageLocationAccessModeReadOnly {
+		_, err := c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+			r.Status.Phase = velerov1api.DeleteBackupRequestPhaseProcessed
 			r.Status.Errors = append(r.Status.Errors, fmt.Sprintf("cannot delete backup because backup storage location %s is currently in read-only mode", location.Name))
 		})
 		return err
@@ -236,11 +252,11 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 	}
 
 	// Update status to InProgress and set backup-name label if needed
-	req, err = c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-		r.Status.Phase = v1.DeleteBackupRequestPhaseInProgress
+	req, err = c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+		r.Status.Phase = velerov1api.DeleteBackupRequestPhaseInProgress
 
-		if req.Labels[v1.BackupNameLabel] == "" {
-			req.Labels[v1.BackupNameLabel] = label.GetValidName(req.Spec.BackupName)
+		if req.Labels[velerov1api.BackupNameLabel] == "" {
+			req.Labels[velerov1api.BackupNameLabel] = label.GetValidName(req.Spec.BackupName)
 		}
 	})
 	if err != nil {
@@ -248,9 +264,9 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 	}
 
 	// Set backup-uid label if needed
-	if req.Labels[v1.BackupUIDLabel] == "" {
-		req, err = c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-			req.Labels[v1.BackupUIDLabel] = string(backup.UID)
+	if req.Labels[velerov1api.BackupUIDLabel] == "" {
+		req, err = c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+			req.Labels[velerov1api.BackupUIDLabel] = string(backup.UID)
 		})
 		if err != nil {
 			return err
@@ -258,15 +274,15 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 	}
 
 	// Set backup status to Deleting
-	backup, err = c.patchBackup(backup, func(b *v1.Backup) {
-		b.Status.Phase = v1.BackupPhaseDeleting
+	backup, err = c.patchBackup(backup, func(b *velerov1api.Backup) {
+		b.Status.Phase = velerov1api.BackupPhaseDeleting
 	})
 	if err != nil {
 		log.WithError(errors.WithStack(err)).Error("Error setting backup phase to deleting")
 		return err
 	}
 
-	backupScheduleName := backup.GetLabels()[v1.ScheduleNameLabel]
+	backupScheduleName := backup.GetLabels()[velerov1api.ScheduleNameLabel]
 	c.metrics.RegisterBackupDeletionAttempt(backupScheduleName)
 
 	var errs []string
@@ -279,6 +295,36 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 		errs = append(errs, err.Error())
 	}
 
+	// Download the tarball
+	backupFile, err := downloadToTempFile(backup.Name, backupStore, log)
+	if err != nil {
+		return errors.Wrap(err, "error downloading backup")
+	}
+	defer closeAndRemoveFile(backupFile, c.logger)
+
+	actions, err := pluginManager.GetDeleteItemActions()
+	log.Debugf("%d actions before invoking actions", len(actions))
+	if err != nil {
+		return errors.Wrap(err, "error getting delete item actions")
+	}
+	// don't defer CleanupClients here, since it was already called above.
+
+	ctx := &delete.Context{
+		Backup:          backup,
+		BackupReader:    backupFile,
+		Actions:         actions,
+		Log:             c.logger,
+		DiscoveryHelper: c.helper,
+		Filesystem:      filesystem.NewFileSystem(),
+	}
+
+	// Optimization: wrap in a gofunc? Would be useful for large backups with lots of objects.
+	// but what do we do with the error returned? We can't just swallow it as that may lead to dangling resources.
+	err = delete.InvokeDeleteActions(ctx)
+	if err != nil {
+		return errors.Wrap(err, "error invoking delete item actions")
+	}
+
 	if backupStore != nil {
 		log.Info("Removing PV snapshots")
 
@@ -320,6 +366,22 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 		}
 	}
 
+	if features.IsEnabled(velerov1api.CSIFeatureFlag) {
+		log.Info("Removing CSI volumesnapshots")
+		if csiErrs := deleteCSIVolumeSnapshots(backup.Name, c.csiSnapshotLister, c.csiSnapshotClient.SnapshotV1beta1(), log); len(csiErrs) > 0 {
+			for _, err := range csiErrs {
+				errs = append(errs, err.Error())
+			}
+		}
+
+		log.Info("Removing CSI volumesnapshotcontents")
+		if csiErrs := deleteCSIVolumeSnapshotContents(backup.Name, c.csiSnapshotContentLister, c.csiSnapshotClient.SnapshotV1beta1(), log); len(csiErrs) > 0 {
+			for _, err := range csiErrs {
+				errs = append(errs, err.Error())
+			}
+		}
+	}
+
 	log.Info("Removing restores")
 	if restores, err := c.restoreLister.Restores(backup.Namespace).List(labels.Everything()); err != nil {
 		log.WithError(errors.WithStack(err)).Error("Error listing restore API objects")
@@ -339,7 +401,7 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 			}
 
 			restoreLog.Info("Deleting restore referencing backup")
-			if err := c.restoreClient.Restores(restore.Namespace).Delete(restore.Name, &metav1.DeleteOptions{}); err != nil {
+			if err := c.restoreClient.Restores(restore.Namespace).Delete(context.TODO(), restore.Name, metav1.DeleteOptions{}); err != nil {
 				errs = append(errs, errors.Wrapf(err, "error deleting restore %s", kube.NamespaceAndName(restore)).Error())
 			}
 		}
@@ -347,7 +409,7 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 
 	if len(errs) == 0 {
 		// Only try to delete the backup object from kube if everything preceding went smoothly
-		err = c.backupClient.Backups(backup.Namespace).Delete(backup.Name, nil)
+		err = c.backupClient.Backups(backup.Namespace).Delete(context.TODO(), backup.Name, metav1.DeleteOptions{})
 		if err != nil {
 			errs = append(errs, errors.Wrapf(err, "error deleting backup %s", kube.NamespaceAndName(backup)).Error())
 		}
@@ -360,8 +422,8 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 	}
 
 	// Update status to processed and record errors
-	req, err = c.patchDeleteBackupRequest(req, func(r *v1.DeleteBackupRequest) {
-		r.Status.Phase = v1.DeleteBackupRequestPhaseProcessed
+	req, err = c.patchDeleteBackupRequest(req, func(r *velerov1api.DeleteBackupRequest) {
+		r.Status.Phase = velerov1api.DeleteBackupRequestPhaseProcessed
 		r.Status.Errors = errs
 	})
 	if err != nil {
@@ -371,7 +433,7 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 	// Everything deleted correctly, so we can delete all DeleteBackupRequests for this backup
 	if len(errs) == 0 {
 		listOptions := pkgbackup.NewDeleteBackupRequestListOptions(backup.Name, string(backup.UID))
-		err = c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).DeleteCollection(nil, listOptions)
+		err = c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).DeleteCollection(context.TODO(), metav1.DeleteOptions{}, listOptions)
 		if err != nil {
 			// If this errors, all we can do is log it.
 			c.logger.WithField("backup", kube.NamespaceAndName(backup)).Error("error deleting all associated DeleteBackupRequests after successfully deleting the backup")
@@ -383,7 +445,7 @@ func (c *backupDeletionController) processRequest(req *v1.DeleteBackupRequest) e
 
 func volumeSnapshotterForSnapshotLocation(
 	namespace, snapshotLocationName string,
-	snapshotLocationLister listers.VolumeSnapshotLocationLister,
+	snapshotLocationLister velerov1listers.VolumeSnapshotLocationLister,
 	pluginManager clientmgmt.Manager,
 ) (velero.VolumeSnapshotter, error) {
 	snapshotLocation, err := snapshotLocationLister.VolumeSnapshotLocations(namespace).Get(snapshotLocationName)
@@ -403,11 +465,9 @@ func volumeSnapshotterForSnapshotLocation(
 	return volumeSnapshotter, nil
 }
 
-func (c *backupDeletionController) deleteExistingDeletionRequests(req *v1.DeleteBackupRequest, log logrus.FieldLogger) []error {
+func (c *backupDeletionController) deleteExistingDeletionRequests(req *velerov1api.DeleteBackupRequest, log logrus.FieldLogger) []error {
 	log.Info("Removing existing deletion requests for backup")
-	selector := labels.SelectorFromSet(labels.Set(map[string]string{
-		v1.BackupNameLabel: label.GetValidName(req.Spec.BackupName),
-	}))
+	selector := label.NewSelectorForBackup(req.Spec.BackupName)
 	dbrs, err := c.deleteBackupRequestLister.DeleteBackupRequests(req.Namespace).List(selector)
 	if err != nil {
 		return []error{errors.Wrap(err, "error listing existing DeleteBackupRequests for backup")}
@@ -419,7 +479,7 @@ func (c *backupDeletionController) deleteExistingDeletionRequests(req *v1.Delete
 			continue
 		}
 
-		if err := c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).Delete(dbr.Name, nil); err != nil {
+		if err := c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).Delete(context.TODO(), dbr.Name, metav1.DeleteOptions{}); err != nil {
 			errs = append(errs, errors.WithStack(err))
 		}
 	}
@@ -427,7 +487,7 @@ func (c *backupDeletionController) deleteExistingDeletionRequests(req *v1.Delete
 	return errs
 }
 
-func (c *backupDeletionController) deleteResticSnapshots(backup *v1.Backup) []error {
+func (c *backupDeletionController) deleteResticSnapshots(backup *velerov1api.Backup) []error {
 	if c.resticMgr == nil {
 		return nil
 	}
@@ -450,6 +510,82 @@ func (c *backupDeletionController) deleteResticSnapshots(backup *v1.Backup) []er
 	return errs
 }
 
+func setVolumeSnapshotContentDeletionPolicy(vscName string, csiClient snapshotter.SnapshotV1beta1Interface, log *logrus.Entry) error {
+	log.Infof("Setting DeletionPolicy of CSI volumesnapshotcontent %s to Delete", vscName)
+	pb := []byte(`{"spec":{"deletionPolicy":"Delete"}}`)
+	_, err := csiClient.VolumeSnapshotContents().Patch(context.TODO(), vscName, types.MergePatchType, pb, metav1.PatchOptions{})
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
+func deleteCSIVolumeSnapshots(backupName string, csiSnapshotLister snapshotv1beta1listers.VolumeSnapshotLister,
+	csiClient snapshotter.SnapshotV1beta1Interface, log *logrus.Entry) []error {
+	errs := []error{}
+
+	selector := label.NewSelectorForBackup(backupName)
+	csiVolSnaps, err := csiSnapshotLister.List(selector)
+	if err != nil {
+		return []error{err}
+	}
+
+	log.Infof("Deleting %d CSI volumesnapshots", len(csiVolSnaps))
+	for _, csiVS := range csiVolSnaps {
+		log.Infof("Deleting CSI volumesnapshot %s/%s", csiVS.Namespace, csiVS.Name)
+		if csiVS.Status != nil && csiVS.Status.BoundVolumeSnapshotContentName != nil {
+			// we patch the DeletionPolicy of the volumesnapshotcontent to set it to Delete.
+			// This ensures that the volume snapshot in the storage provider is also deleted.
+			err := setVolumeSnapshotContentDeletionPolicy(*csiVS.Status.BoundVolumeSnapshotContentName, csiClient, log)
+			if err != nil && !apierrors.IsNotFound(err) {
+				log.Errorf("Skipping deletion of volumesnapshot %s/%s", csiVS.Namespace, csiVS.Name)
+				errs = append(errs, err)
+				continue
+			}
+		}
+		err := csiClient.VolumeSnapshots(csiVS.Namespace).Delete(context.TODO(), csiVS.Name, metav1.DeleteOptions{})
+		if err != nil {
+			errs = append(errs, err)
+		}
+	}
+	return errs
+}
+
+func deleteCSIVolumeSnapshotContents(backupName string, csiVSCLister snapshotv1beta1listers.VolumeSnapshotContentLister,
+	csiClient snapshotter.SnapshotV1beta1Interface, log *logrus.Entry) []error {
+	errs := []error{}
+	selector := label.NewSelectorForBackup(backupName)
+	csiVolSnapConts, err := csiVSCLister.List(selector)
+	if err != nil {
+		return []error{err}
+	}
+	// It is possible that by the time deleteCSIVolumeSnapshotContents is called after deleteCSIVolumeSnapshots
+	// that deletion of VSCs hasn't been completed, by the snapshot-controller (one of the CSI components).
+	// For that reason the csiVSCLister returned VSCs that are yet to be deleted. To handle this scenario,
+	// we swallow `IsNotFound` errors from the setVolumeSnapshotContentDeletionPolicy function and the
+	// csiClient.VolumeSnapshotContents().Delete(...)
+	log.Infof("Deleting %d CSI volumesnapshotcontents", len(csiVolSnapConts))
+	for _, snapCont := range csiVolSnapConts {
+		err := setVolumeSnapshotContentDeletionPolicy(snapCont.Name, csiClient, log)
+		if err != nil && !apierrors.IsNotFound(err) {
+			log.Errorf("Failed to set DeletionPolicy on volumesnapshotcontent %s. Skipping deletion", snapCont.Name)
+			errs = append(errs, err)
+			continue
+		}
+		if apierrors.IsNotFound(err) {
+			log.Infof("volumesnapshotcontent %s not found", snapCont.Name)
+			continue
+		}
+		log.Infof("Deleting volumesnapshotcontent %s", snapCont.Name)
+		err = csiClient.VolumeSnapshotContents().Delete(context.TODO(), snapCont.Name, metav1.DeleteOptions{})
+		if err != nil && !apierrors.IsNotFound(err) {
+			errs = append(errs, err)
+		}
+	}
+
+	return errs
+}
+
 const deleteBackupRequestMaxAge = 24 * time.Hour
 
 func (c *backupDeletionController) deleteExpiredRequests() {
@@ -466,7 +602,7 @@ func (c *backupDeletionController) deleteExpiredRequests() {
 	now := c.clock.Now()
 
 	for _, req := range requests {
-		if req.Status.Phase != v1.DeleteBackupRequestPhaseProcessed {
+		if req.Status.Phase != velerov1api.DeleteBackupRequestPhaseProcessed {
 			continue
 		}
 
@@ -475,7 +611,7 @@ func (c *backupDeletionController) deleteExpiredRequests() {
 			reqLog := c.logger.WithFields(logrus.Fields{"namespace": req.Namespace, "name": req.Name})
 			reqLog.Info("Deleting expired DeleteBackupRequest")
 
-			err = c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).Delete(req.Name, nil)
+			err = c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).Delete(context.TODO(), req.Name, metav1.DeleteOptions{})
 			if err != nil {
 				reqLog.WithError(err).Error("Error deleting DeleteBackupRequest")
 			}
@@ -483,7 +619,7 @@ func (c *backupDeletionController) deleteExpiredRequests() {
 	}
 }
 
-func (c *backupDeletionController) patchDeleteBackupRequest(req *v1.DeleteBackupRequest, mutate func(*v1.DeleteBackupRequest)) (*v1.DeleteBackupRequest, error) {
+func (c *backupDeletionController) patchDeleteBackupRequest(req *velerov1api.DeleteBackupRequest, mutate func(*velerov1api.DeleteBackupRequest)) (*velerov1api.DeleteBackupRequest, error) {
 	// Record original json
 	oldData, err := json.Marshal(req)
 	if err != nil {
@@ -504,7 +640,7 @@ func (c *backupDeletionController) patchDeleteBackupRequest(req *v1.DeleteBackup
 		return nil, errors.Wrap(err, "error creating json merge patch for DeleteBackupRequest")
 	}
 
-	req, err = c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).Patch(req.Name, types.MergePatchType, patchBytes)
+	req, err = c.deleteBackupRequestClient.DeleteBackupRequests(req.Namespace).Patch(context.TODO(), req.Name, types.MergePatchType, patchBytes, metav1.PatchOptions{})
 	if err != nil {
 		return nil, errors.Wrap(err, "error patching DeleteBackupRequest")
 	}
@@ -512,7 +648,7 @@ func (c *backupDeletionController) patchDeleteBackupRequest(req *v1.DeleteBackup
 	return req, nil
 }
 
-func (c *backupDeletionController) patchBackup(backup *v1.Backup, mutate func(*v1.Backup)) (*v1.Backup, error) {
+func (c *backupDeletionController) patchBackup(backup *velerov1api.Backup, mutate func(*velerov1api.Backup)) (*velerov1api.Backup, error) {
 	// Record original json
 	oldData, err := json.Marshal(backup)
 	if err != nil {
@@ -533,7 +669,7 @@ func (c *backupDeletionController) patchBackup(backup *v1.Backup, mutate func(*v
 		return nil, errors.Wrap(err, "error creating json merge patch for Backup")
 	}
 
-	backup, err = c.backupClient.Backups(backup.Namespace).Patch(backup.Name, types.MergePatchType, patchBytes)
+	backup, err = c.backupClient.Backups(backup.Namespace).Patch(context.TODO(), backup.Name, types.MergePatchType, patchBytes, metav1.PatchOptions{})
 	if err != nil {
 		return nil, errors.Wrap(err, "error patching Backup")
 	}
